/*
 *  Flash Rom Disk Driver
 *
 *  Copyright (C) 2001-2004  BUFFALO INC.
 *
 *  This software may be used and distributed according to the terms of
 *  the GNU General Public License (GPL), incorporated herein by reference.
 *  Drivers based on or derived from this code fall under the GPL and must
 *  retain the authorship, copyright and license notice.  This file is not
 *  a complete program and may only be used when the entire operating
 *  system is licensed under the GPL.
 *
 */
#ifndef _H_FLASH_TOSHIBA_
#define _H_FLASHD_TOSHIBA_

#include <melco/melco_hwctl.h>

static int flash_amd_compt_probe(unsigned char factory , unsigned char devnum);
static int flash_amd_compt_erase(unsigned long addr ,int num , int flag);

static int flash_toshiba_probe(void);
static int flash_toshiba_erase(unsigned long addr ,int num);
static int flash_toshiba_write(unsigned long addr ,unsigned char* buf,int size);

static int flash_fujitsu_probe(void);
static int flash_fujitsu_erase(unsigned long addr ,int num);

#define HEAD_BLOCK_8K    0x01
#define BOTTOM_BLOCK_8K  0x02

struct flash_dev_func flash_toshiba  = 
	{
	"TOSHIBA VT641FT",
	64 * 1024,
	flash_toshiba_probe,
	flash_toshiba_erase,
	flash_toshiba_write,
	NULL
	};

struct flash_dev_func flash_fujitsu  = 
	{
	"FUJITSU 29DL640E",
	64 * 1024,
	flash_fujitsu_probe,
	flash_fujitsu_erase,
	flash_toshiba_write,
	NULL
	};

#define TOSHIBA_FLASH_CMDA(x)	\
	*(volatile char*)(ATTACH_FLASH_ADDRESS_1 + 0xaaa) = x;\
	*(volatile char*)(ATTACH_FLASH_ADDRESS_2 + 0xaaa) = x;
#define TOSHIBA_FLASH_CMD5(x)	\
	*(volatile char*)(ATTACH_FLASH_ADDRESS_1 + 0x555) = x;\
	*(volatile char*)(ATTACH_FLASH_ADDRESS_2 + 0x555) = x;
#define TOSHIBA_FLASH_START (volatile char*)(ATTACH_FLASH_ADDRESS_1)

#define FLASH_COMMAND_CLEAR TOSHIBA_FLASH_CMDA(0xf0)
#define FLASH_COMMAND_READY TOSHIBA_FLASH_CMDA(0xaa);TOSHIBA_FLASH_CMD5(0x55)

static int flash_toshiba_probe(void) 
{
	return flash_amd_compt_probe(0x98,0x93);
}

static int flash_toshiba_erase(unsigned long addr ,int num) 
{
	return flash_amd_compt_erase(addr ,num , BOTTOM_BLOCK_8K);
};

static int flash_fujitsu_probe(void) 
{
	return flash_amd_compt_probe(0x04,0x7e);
}

static int flash_fujitsu_erase(unsigned long addr ,int num) {
	return flash_amd_compt_erase(addr ,num , HEAD_BLOCK_8K | BOTTOM_BLOCK_8K);
	};

static int flash_amd_compt_probe(unsigned char factory , unsigned char devnum)
{
	unsigned long time = jiffies;
	unsigned char vendor,device;
	FLASH_WRITE_ENABLE;
	FLASH_COMMAND_CLEAR;
	FLASH_COMMAND_READY;
	TOSHIBA_FLASH_CMDA(0x90);
	do {
		vendor = *(TOSHIBA_FLASH_START);
		device = *(TOSHIBA_FLASH_START+2);
		if(vendor == factory || device == devnum) break;
		} while(jiffies-time<500);
	FLASH_COMMAND_CLEAR;
	FLASH_WRITE_DISABLE;
	if(vendor == factory || device == devnum) return 1;
	return 0;
}

#define BIT(x) 1<<x
static int flash_toshiba_state(volatile unsigned char* addr ,unsigned char target)
{
	unsigned long time = jiffies;
	unsigned char data;
	do {
		data = *addr;
		if((data & BIT(7))==(target & BIT(7))) return 0;
		if(data & BIT(5)) {
			data = *addr;
			if((data & BIT(7))==(target & BIT(7))) 
				return 0;
			else 
				return -1;
			}
		} while(jiffies-time<5000);
	return -1;
}

static int flash_amd_compt_erase(unsigned long addr ,int num , int flag)
{
	volatile unsigned char* baddr;
	volatile unsigned char* caddr;
	unsigned char           c;
	long  i,times;
	long  size;
	addr  = addr & 0xffff0000;
	baddr = (char*)addr;

	while(num>0) {
		if( ((flag & HEAD_BLOCK_8K) != 0) &&
			(((unsigned long)baddr >= (unsigned long)0xff000000 &&
			  (unsigned long)baddr < (unsigned long)0xff010000) ||
			 ((unsigned long)baddr >= (unsigned long)0xff800000 &&
			  (unsigned long)baddr < (unsigned long)0xff810000)) )
		{
				times = 8;
				size  = 0x2000;
		}
		else if( ((flag & BOTTOM_BLOCK_8K) != 0) &&
			(((unsigned long)baddr >= (unsigned long)0xff7f0000 &&
			  (unsigned long)baddr < (unsigned long)0xff800000) ||
			 ((unsigned long)baddr >= (unsigned long)0xffff0000)) )
		{
			times = 8;
			size  = 0x2000;
		}
		else {
			times = 1;
			size  = 0x10000;
			}
		for(i=0;i<times;++i) {
			FLASH_WRITE_ENABLE;
			FLASH_COMMAND_CLEAR;
			FLASH_COMMAND_READY;
			TOSHIBA_FLASH_CMDA(0x80);
			FLASH_COMMAND_READY;
			*(baddr+i*size) = 0x30;
			if(flash_toshiba_state(baddr+i*size,0xff)) {
				printk("Flash Erase fail?!\n");
			}
			FLASH_COMMAND_CLEAR;
			FLASH_WRITE_DISABLE;
			}
		caddr = baddr;
		for(i=0;i<0x10000;++i) {
			c = *caddr;
			if(c != (unsigned char)0xff) {
				printk("verify fail! %p = %02X\n",caddr,c);
				return -1;
				}
			caddr++;
			}
		num--;
		baddr += 0x10000;
		}

	return 0;
}

static int flash_toshiba_write(unsigned long addr ,unsigned char* buf,int size)
{
	volatile unsigned char* baddr;
	long                    i;
	int                     is_blink = 0;
	baddr = (char*)addr;

	FLASH_WRITE_ENABLE;
	FLASH_COMMAND_CLEAR;
	FLASH_COMMAND_READY;
	TOSHIBA_FLASH_CMDA(0x20);
	for(i=0;i<size;++i) {
		*baddr = 0xa0;
		*baddr = buf[i];
		if(flash_toshiba_state(baddr,buf[i])) {
			printk("Flash Write fail?!\n");
			}
		baddr++;
		}
	TOSHIBA_FLASH_CMDA(0x90);
	FLASH_COMMAND_CLEAR;
	FLASH_WRITE_DISABLE;
	baddr = (char*)addr;
	for(i=0;i<size;++i) {
		if(baddr[i] != buf[i]) return -1;
		}

	return 0;
}

#endif
